#!/usr/bin/env python
""" Tools to manage Properties attached to client classes
"""

import time

from collections import OrderedDict as odict


from .utils import cast_type, Meta, Defs, defaults_decorator, defaults_docstring

try:
    basestring
except NameError:
    basestring = str



class Property(Defs):
    """Base class to attach managed attribute to class.

    Notes
    -----
    This is a "validator" class:  https://docs.python.org/3/howto/descriptor.html#validator-class

    It manages a Property for a client class using the __set_name__, __set__, __get__, and __delete__ functions.

    In the example above, for an object my_obj of client class MyClass
    the property val_float would store a hidden attribute my_obj._val_float

    my_obj.val_float uses Property.__get__() to access my_obj._val_float
    my_obj.val_float = 3.3 uses Property.__set__() to set my_obj._val_float (and validate the input value)
    del my_obj.val_float uses Property.__delete__() to set my_obj._val_float the default value
    """
    __metaclass__ = Meta

    defaults = odict([
        ('help', ("", 'Help description')),
        ('format', ('%s', 'Format string for printing')),
        ('dtype', (None, 'Data type')),
        ('default', (None, 'Default value')),
        ('required', (False, 'Is this propery required?')),
    ])

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        self.help = None
        self.format = None
        self.dtype = type(None)
        self.default = None
        self.required = None
        self.public_name = None
        self.private_name = None
        self.time_name = None
        super(Property, self).__init__()
        self._load(**kwargs)

    def __set_name__(self, owner, name):
        """Set the name of the privately managed value"""
        self.public_name = name
        self.private_name = '_' + name
        self.time_name = '_' + name + '_timestamp'


    def __set__(self, obj, value):
        """Set the value in the client object

        Parameter
        ---------
        obj : ...
            The client object
        value : ...
            The value being set
        This will use the `cast_type(self.dtype, value)` method to cast the requested value to the correct type.

        Rasies
        ------
        TypeError : The input value is the wrong type (i.e., not castable to self.dtype)

        ValueError : The input value failes validation for a Property sub-class (e.g., not a valid choice, or outside bounds)
        """
        try:
            cast_value = self._cast_type(value, obj)
            self.validate_value(obj, cast_value)
        except (TypeError, ValueError) as msg:
            setattr(obj, self.private_name, None)
            setattr(obj, self.time_name, time.time())
            raise TypeError("Failed to set %s %s %s" % (repr(obj), self.private_name, msg)) from msg
        setattr(obj, self.private_name, cast_value)
        setattr(obj, self.time_name, time.time())
        if hasattr(obj, '_timestamp'):
            setattr(obj, '_timestamp', time.time())

    def __get__(self, obj, objtype=None):
        """Get the value from the client object

        Parameter
        ---------
        obj : ...
            The client object

        Return
        ------
        out : ...
            The requested value
        """
        return getattr(obj, self.private_name)

    def __delete__(self, obj):
        """Set the value to the default value

        This can be useful for sub-classes that use None
        to indicate an un-initialized value.
        """
        setattr(obj, self.private_name, self.default)
        setattr(obj, self.time_name, time.time())
        if hasattr(obj, '_timestamp'):
            setattr(obj, '_timestamp', time.time())

    def _load(self, **kwargs):
        """Load kwargs key,value pairs into __dict__
        """
        defaults = {}
        # Require kwargs are in defaults
        for k in kwargs:
            if k not in self.defaults:
                msg = "Unrecognized attribute of %s: %s" % (self.__class__.__name__, k)
                raise AttributeError(msg)

        defaults.update(kwargs)

        # This doesn't overwrite the properties
        self.__dict__.update(defaults)

        # Make sure the default is valid
        _ = self._cast_type(self.default)


    def _cast_type(self, value, obj=None): #pylint: disable=unused-argument
        """Hook took override type casting"""
        return cast_type(self.dtype, value)

    def timestamp(self, obj):
        """Get the timestamp when this object was updated"""
        return max(getattr(obj, self.time_name), getattr(getattr(obj, self.private_name), '_timestamp', 0))

    @classmethod
    def defaults_docstring(cls, header=None, indent=None, footer=None):
        """Add the default values to the class docstring"""
        return defaults_docstring(cls.defaults, header=header, indent=indent, footer=footer) #pragma: no cover

    def validate_value(self, obj, value): #pylint: disable=unused-argument,no-self-use
        """Validate a value"""
        return

    def todict(self, obj):
        """Extract value """
        val = getattr(obj, self.private_name)
        if hasattr(val, 'todict'):
            return val.todict()
        return val



#!/usr/bin/env python
"""
Base class for configurable objects
"""
from collections import OrderedDict as odict
from collections.abc import Mapping

from .property import Property
from .derived import Derived
from .ref import Ref

from .utils import Meta, model_docstring

class Configurable:
    """Base class for Configurable objects

    Examples::

        # A simple class with some managed Properties
        class MyClass(Configurable):

            val_float = Property(dtype=float, default=0.0, help="A float value")
            val_int = Property(dtype=int, default=1, help="An integer value")
            val_required = Property(dtype=float, required=True, help="A float value")
            val_str = Property(dtype=str, default="", help="A string value")
            val_list = Property(dtype=list, default=[], help="An list value")
            val_dict = Property(dtype=list, default={}, help="A dictionary value")

        # A class with nested configuration
        class MyPair(Model):
            val_first = Property(dtype=MyClass, required=True, help="First MyClass object")
            val_second = Property(dtype=MyClass, required=True, help="Second MyClass object")


        # Default, all Properties take their default values (must provide required Properties)
        my_obj = MyClass(val_required=3.)

        # Access Properties
        my_obj.val_float
        my_obj.val_str
        my_obj.val_list
        my_obj.val_dict
        my_obj.val_required

        # Set Properties
        my_obj.val_float = 5.4
        my_obj.val_int = 3
        my_obj.val_dict = dict(a=3, b=4)

        # This will fail with a TypeError
        my_obj.val_float = "not a float"

        # Override values in construction
        my_obj = MyClass(val_required=3., val_float=4.3, val_int=2, val_str="Hello World")

        # Build nested Configurables
        my_pair = MyPair(val_first=dict(val_required=3.), val_second=dict(val_required=4.))

        my_pair.val_first.val_float
        my_pair.val_second.val_int

    """

    __metaclass__ = Meta

    def __init__(self, **kwargs):
        """ C'tor.  Build from a set of keyword arguments.
        """
        self._timestamp = 0
        self._init_properties(**kwargs)

    @classmethod
    def find_properties(cls):
        """Find properties that belong to this model"""
        the_classes = cls.mro()
        props = odict()
        for the_class in the_classes:
            for key, val in the_class.__dict__.items():
                if isinstance(val, Property):
                    props[key] = val
        return props

    @classmethod
    def getp(cls, name):
        """Get a particular property of this model"""
        props = cls.find_properties()
        return props[name]

    @property
    def _properties(self):
        """Return the properties of this class"""
        return self.find_properties()

    def __str__(self, indent=0):
        """ Cast model as a formatted string
        """
        props = self._properties
        try:
            ret = '{0:>{2}}{1}'.format('', self.name, indent)
        except AttributeError:
            ret = "%s" % (type(self))
        if not props: #pragma: no cover
            return ret
        ret += '\n{0:>{2}}{1}'.format('', 'Parameters:', indent + 2)

        width = len(max(props.keys(), key=len))
        for prop in props.values():
            name = prop.public_name
            if isinstance(prop, Derived):
                value = getattr(self, prop.private_name)
            else:
                value = getattr(self, name)
            par = '{0!s:{width}} : {1!r}'.format(name, value, width=width)
            ret += '\n{0:>{2}}{1}'.format('', par, indent + 4)
        return ret

    @classmethod
    def defaults_docstring(cls, header=None, indent=None, footer=None):
        """Add the default values to the class docstring"""
        return model_docstring(cls, header=header, indent=indent, footer=footer) #pragma: no cover

    def update(self, *args, **kwargs):
        """
        Set a group of attributes (parameters and members).  Calls
        `setp` directly, so kwargs can include more than just the
        parameter value (e.g., bounds, free, etc.).
        """
        if args:
            raise ValueError("Argument passed to Model.upate() %s" % args)

        kwargs = dict(kwargs)
        props = self._properties
        for name, value in kwargs.items():
            # Raise KeyError if Property not found
            try:
                prop = props[name]
            except KeyError as err:
                raise KeyError("Warning: %s does not have properties %s" % (type(self), name)) from err

            if isinstance(prop, (Ref, Derived)):
                continue
            attr = getattr(self, '_%s' % name)
            if isinstance(attr, Configurable):
                # Set attributes
                if isinstance(value, Mapping):
                    attr.update(**value)
                else:
                    attr.update(value)
            else:
                prop.__set__(self, value)


    def _init_properties(self, **kwargs):
        """ Loop through the list of Properties,
        extract the derived and required properties and do the
        appropriate book-keeping
        """
        missing = {}
        kwcopy = kwargs.copy()
        props = self._properties
        for k, p in props.items():
            if k not in kwcopy and p.required:
                missing[k] = p

            pval = kwcopy.pop(k, p.default)
            if isinstance(p, Ref):
                setattr(self, p.owner_name, p.owner)
                setattr(self, p.private_name, None)
                continue
            p.__set__(self, pval)
        if missing:
            raise ValueError("One or more required properties are missing ", type(self), missing.keys())
        if kwcopy:
            raise KeyError("Warning: %s does not have attributes %s" % (type(self), kwcopy))

    def todict(self):
        """ Return self cast as an '~collections.OrderedDict' object
        """
        return odict([(key, val.todict(self)) for key, val in self._properties.items()])



"""Small module for unit converision"""

import numpy as np

class Unit:
    """
    Object for handling unit conversions

    """
    to_SI_dict = {}

    def __init__(self, unit=None):
        # Dictionary of SI unit conversions
        # Check that passed unit is available
        if unit is None:
            self._SI = 1.
            self._name = ''
            return
        if isinstance(unit, str):
            if unit not in self.to_SI_dict:
                raise KeyError("Passed unit '%s' not understood by Unit object" % (unit))
            self._SI = self.to_SI_dict[unit]
            self._name = unit
            return
        self._SI = float(unit)
        self._name = "a.u."

    @property
    def name(self):
        """Return the units name"""
        return self._name

    def __call__(self, val):
        """Convert value to SI unit """
        if val is None:
            return None
        return np.array(val) * self._SI

    def inverse(self, val):
        """Convert value from SI unit """
        if val is None:
            return None
        return np.array(val) / self._SI

    @classmethod
    def update(cls, a_dict):
        """Update the mapping of unit names"""
        cls.to_SI_dict.update(a_dict)



"""Tools to build classes"""

from collections import OrderedDict as odict

from .utils import expand_dict
from .property import Property

def build_sub_configurables(config_dict, type_dict):
    """ Build a set of of Property objects to manage a set of Configurables,

    Parameters
    ----------
    config_dict : `dict`
        Dictionary used to build the sub-objects

    type_dict : `dict`
        Dictionary mapping names to Configurable Class names

    Returns
    -------
    o_dict : `dict`
        odict(name:Property), i.e., a dictionary mapping names to the Property object to manage the owned objects

    Notes
    -----
    This function is used by build_class to extend classes algorithmically by attaching Property objects that will be used to
    access the data members of the class.
    """

    return odict([(key, Property(dtype=type_dict[val.pop('obj_type', None)])) for key, val in config_dict.items()])


def build_class(name, base_classes, config_dicts, type_dicts, **kwargs):
    """ Build a new class algorithmically.

    Parameters
    ----------
    name : `str`
        The name of the new class

    base_classes : `Iterable`
        The base classes

    config_dicts : `Iterable`
        Dictionaries used to build the sub-objects

    type_dicts : `Iterable`
        Dictionaries mapping names to Configurable Class names for each of the config_dicts


    Returns
    -------
    object : `Configurable`
        Newly-made object of the newly made type.
        The kwargs are based directly to the constructor of the new class, and the resulting object is returned.

    Notes
    -----
    This function is used to extend classes `Configurable` sub-classes algorithmically.

    If a `Configurable` object has some optional data members, this class will build them based
    on the config_dict.
    """
    kwcopy = kwargs.copy()

    props = odict()
    for config_dict_, type_dict_ in zip(config_dicts, type_dicts):
        use_dict_ = expand_dict(config_dict_)
        kwcopy.update(use_dict_)
        props.update(build_sub_configurables(use_dict_, type_dict_))

    new_class = type(name, base_classes, props)
    return new_class(**kwcopy)



#!/usr/bin/env python
"""
A few simple utilities to help parse configurations
"""
from collections import OrderedDict as odict
from collections.abc import Mapping

import yaml

try:
    basestring
except NameError:
    basestring = str


def expand_dict(in_dict):
    """Expand a dictionary by copying defaults to a set of elements

    Parameters
    ----------
    in_dict : `dict`
        The input dictionary

    elem_dict : `dict`
        The elements

    Returns
    -------
    o_dict : `dict`
        The output dict
    """
    if 'default' not in in_dict:
        return in_dict
    default_dict = in_dict.get('default')
    elem_dict = in_dict.get('elements')
    o_dict = odict()
    for key, elem in elem_dict.items():
        o_dict[key] = default_dict.copy()
        if elem is None:
            continue
        o_dict[key].update(elem)
    return o_dict




def defaults_docstring(defaults, header='', indent='', footer=''):
    """Return a docstring from a list of defaults.
    """

    #width = 60
    #hbar = indent + width * '=' + '\n'  # horizontal bar
    hbar = '\n'

    s = hbar + (header) + hbar
    for key, (value, desc) in defaults.items():
        if isinstance(value, basestring):
            value = "'" + value + "'"
        if hasattr(value, '__call__'):
            value = "<" + value.__name__ + ">"

        s += indent +'%-12s\n' % ("%s :" % key)
        s += indent + indent + (indent + 23 * ' ').join(desc.split('\n'))
        s += ' [%s]\n\n' % str(value)
    s += hbar
    s += footer
    return s

def defaults_decorator(defaults):
    """Decorator to append default kwargs to a function.
    """
    def decorator(func):
        """Function that appends default kwargs to a function.
        """
        kwargs = dict(header='Keyword arguments\n-----------------\n',
                      indent='  ',
                      footer='\n')
        doc = defaults_docstring(defaults, **kwargs)
        if func.__doc__ is None:
            func.__doc__ = ''
        func.__doc__ += doc
        return func

    return decorator


def model_docstring(cls, header='', indent='', footer=''): #pragma: no cover
    """Return a docstring from a list of defaults.
    """
    #width = 60
    #hbar = indent + width * '=' + '\n'  # horizontal bar
    hbar = '\n'

    props = cls.find_properties()

    s = hbar + (header) + hbar
    for key, val in props.items():
        s += indent +'%-12s\n' % ("%s :" % key)
        s += indent + indent + (indent + 23 * ' ').join(val.help.split('\n'))
        s += ' [%s]\n\n' % str(val.default)
    s += hbar
    s += footer
    return s


class Meta(type): #pragma: no cover
    """Meta class for appending docstring with defaults
    """
    def __new__(mcs, name, bases, attrs):
        attrs['_doc'] = attrs.get('__doc__', '')
        return super(Meta, mcs).__new__(mcs, name, bases, attrs)

    @property
    def __doc__(cls):
        kwargs = dict(header='Parameters\n----------\n',
                      indent='  ',
                      footer='\n')
        return cls._doc + cls.defaults_docstring(**kwargs)


class Defs:
    """ Mixin class to handle default value construction"""
    __metaclass__ = Meta

    defaults = odict()

    def __init__(self, default_prefix=''):
        """Load kwargs key,value pairs into __dict__
        """
        self._default_prefix = default_prefix
        for key, val in self.defaults.items():
            self.__dict__["%s%s" % (default_prefix, key)] = val[0]

    @property
    def default_prefix(self):
        """Get the prefix associated to defaults to make attributes"""
        return self._default_prefix

    def default_value(self, name):
        """Get the default value of a particular attribute"""
        return self.__class__.defaults.get(name)[0]


def is_none(val):
    """Check for values equivalent to None

    This will return True if val is one of None, 'none', 'None'
    """
    if not isinstance(val, (type(None), str)):
        return False
    return val in [None, 'none', 'None']

def is_not_none(val):
    """Check for values equivalent to None

    This will return True if val is not on of  None, 'none', 'None'
    """
    if not isinstance(val, (type(None), str)):
        return True
    return val not in [None, 'none', 'None']


def cast_type(dtype, value): #pylint: disable=too-many-return-statements
    """Casts an input value to a particular type

    Parameters
    ----------
    dtype : type
        The type we are casting to
    value : ...
        The value being cast

    Returns
    -------
    out : ...
        The object cast to dtype

    Raises
    ------
    TypeError if neither value nor dtype are None and the casting fails

    Notes
    -----
    This will proceed in the following order
        1.  If dtype is None it will simply return value
        2.  If value is None, 'None', or 'none' it will return None
        3.  If value is an instance of dtype it will return value
        4.  It will try to pass value to the constructor of dtype, i.e., return dtype(value)
        5.  It will try to use value as a list of argument to the constructor of dtype, i.e., return dtype(*value)
        6.  It will try to use value as a keyword dictionary to the constructor of dtype, i.e., return dtype(**value)
        7.  It will try to extract value['value'] and pass that to the constructor of dtype, i.e., return dtype(value['value'])
    """
    if is_none(dtype):
        return value
    # value = None is always allowed
    if is_none(value):
        return None
    # if value is an instance of self.dtype, then return it
    if isinstance(value, dtype):
        return value
    if isinstance(value, Mapping):
        return dtype(**value)
    if isinstance(value, str):
        return dtype(value)
    try:
        return dtype(value)
    except (TypeError, ValueError):
        pass

    # try and cast value itself to dtype constructor
    #try:
    #    return dtype(value)
    #except (TypeError, ValueError):
    #    pass
    # try and cast the value as a list to dtype constructor
    #try:
    #    return dtype(*value)
    #except (TypeError, ValueError):
    #    pass
    # try and cast the value as a dict to dtype constructor
    #try:
    #    return dtype(**value)
    #except (TypeError, ValueError):
    #    pass
    # try and cast extract the 'value' item from a dict
    #try:
    #    return dtype(value['value'])
    #except (TypeError, ValueError, KeyError):
    #    pass
    msg = "Value of type %s, when %s was expected." % (type(value), dtype)
    raise TypeError(msg)



def odict_representer(dumper, data):
    """ http://stackoverflow.com/a/21912744/4075339 """
    # Probably belongs in a util
    return dumper.represent_mapping(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, data.items())

yaml.add_representer(odict, odict_representer)



""" Tools for configuration parsing and model building """

from .version import get_git_version
__version__ = get_git_version()
del get_git_version

from .unit import Unit
from .ref import Ref
from .array import Array
from .property import Property
from .derived import Derived, cached
from .configurable import Configurable
from .choice import Choice
from .parameter import Parameter
from .param_holder import ParamHolder
from .model import Model

from . import tools, utils



#!/usr/bin/env python
""" Tools to manage Property objects that are derived from other Property objects manages by the same client class.
"""

import numpy as np
from copy import deepcopy

from .property import Property, defaults_decorator

class Derived(Property):
    """Property sub-class for derived configuration Properties (i.e., Properties
    that depend on other Properties)

    This allows specifying the specifying a 'loader' function by name
    that is used to compute the value of the property.
    """
    defaults = deepcopy(Property.defaults)
    defaults['loadername'] = (None, 'Name of function to load datum')
    defaults['loader'] = (None, 'Function to load datatum')
    defaults['uses'] = ([], 'Properties used by this one')

    @classmethod
    def dummy(cls): #pragma: no cover
        """Dummy function"""
        return

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        self.loader = None
        self.loadername = None
        self.uses = []
        super(Derived, self).__init__(**kwargs)

    def __set_name__(self, owner, name):
        """Set the name of the privately managed value"""
        super(Derived, self).__set_name__(owner, name)
        if self.loadername is None:
            if self.loader is not None:
                self.loadername = "_load_%s" % self.loader.__name__
            else:
                self.loadername = "_load_%s" % name
        if self.loader is None:
            loader = getattr(owner, self.loadername, None)
            self.loader = loader
            if not callable(self.loader):
                raise ValueError("Callable loader not defined for Derived object", owner, name, self.loadername, self.loader)
            #return
        setattr(owner, self.loadername, self.loader)

    def __get__(self, obj, objtype=None):
        """Get the value from the client object

        Parameter
        ---------
        obj : ...
            The client object

        Return
        ------
        val : ...
            The requested value

        Notes
        -----
        This first checks if the value is cached (i.e., if getattr(obj, self.private_name) is None)

        If it is not cached then it invokes the `self.loader` function to compute the value, and caches the computed value
        """
        val = getattr(obj, self.private_name)

        loader = self.loader

        if loader is None: #pragma: no cover
            raise ValueError("%s.%s no loader" % (obj, self.public_name))

        ts_check = self.check_timestamp(obj)
        if ts_check and val is not None:
            return val
        val = loader(obj)
        self.__set__(obj, val)
        return getattr(obj, self.private_name)

    def check_timestamp(self, obj):
        """Check if object is up to date w.r.t. to the object it uses"""
        times = [vn.timestamp(obj) for vn in self.uses]
        if times:
            ts_max = np.max(times)
        else:
            ts_max = 0.
        my_ts = self.timestamp(obj)
        return my_ts >= ts_max



def cached(**kwargs):
    """Decorator attach a function to a class as a Derived property
    """
    def decorator(func):
        """Function that appends default kwargs to a function.
        """
        return Derived(loader=func, **kwargs)
    return decorator



#!/usr/bin/env python
""" Tools to manage Property objects that consist of a set of allowed choics
"""

from copy import deepcopy

from .property import Property, defaults_decorator


class Choice(Property):
    """Property sub-class for defining a numerical Parameter.

    This includes value, bounds, error estimates and fixed/free status
    (i.e., for fitting)

    """

    # Better to keep the structure consistent with Property
    defaults = deepcopy(Property.defaults)
    defaults['choices'] = ([], 'Allowed values')

    # Overwrite the default dtype
    defaults['dtype'] = (str, 'Data type')

    @defaults_decorator(defaults)
    def __init__(self, **kwargs):
        super(Choice, self).__init__(**kwargs)

    def validate_value(self, obj, value):
        """Validate a value

        In this case it checks that value is in the set of allowed choices
        """
        if value is None:
            return
        if value not in self.choices: #pylint: disable=no-member
            raise ValueError("%s not allow, options are %s" % (value, self.choices))  #pylint: disable=no-member
